extends Node


var path_ht = {}
#var path_ht_r = {}
#var is_client
var is_server
var is_primary
var is_proxy = false
var initialized = false
#var connecting = false
#var connection_inprogress = false
var connection_pending = false
var connection_waiting = []
var connection_queue = []
var clhost = null
var srvhost = null
var hosted = false;
var connections = {}
var primary_server = 1
var proxy = 1
var servers = []
var server_simulation_delay = 0 # 150# 100

var cnx_tokens = {}
var token_cid = {}


var master_nodes = {}

var uid = -1
var my_primary_addr


var maxplayer = 32;

var pings = {}
var sync_table = {}

var connect_yield
var setup_complete = false

var process_data_queue = []
var send_data_queue = []
var reverse_lag = false

var data_log = []

var replay_data = []
var replay_current_time = 0
var replay_local_time = 0
var replay_pause_time = 0
const REPLAY_STOPPED = 0
const REPLAY_PLAYING = 1
const REPLAY_PAUSED = 2
var replay_state = REPLAY_STOPPED



onready var g = get_node("/root/global")
onready var gsvar = get_node("/root/gamestate")

signal client_connected(pid)
signal client_disconnected(pid)
signal connected_to_server(pid)
signal connection_failed(pid)
signal server_disconnected(pid)
signal latency_change(pid, diff, direct)
signal done_connecting()


class NetActor:
  var connected = false;
  var initializing = false
  var fail = false
  var peer
  var address = []
  var token



#const NET_CLIENT_CONNECTED = 1
#
#const NET_CLIENT_DISCONNECTED = 3
const NET_ACCEPTED = 4

const NET_RPC = 10
const NET_RSET = 11


func _ready():
#  connections.clear();
#  connections.resize(maxplayer);
  Globals["net"] = self
  for i in range(1, maxplayer+1):
    connections[i] = {"srv":NetActor.new(), "cl":NetActor.new()};
  self.connect("done_connecting", self, "start_auto_proxy")
  setup_complete = true
  set_process(true);

#func _notification(what):
  #if (what == MainLoop.NOTIFICATION_WM_QUIT_REQUEST or what == Node.NOTIFICATION_EXIT_TREE ):
#    print(["Shutting down network", what])
#    close_connections()
##    get_tree().quit()


func reset():
  is_server = false
  servers = []
  clhost = null
  srvhost = null
  _ready()


func close_connections():
  print("Closing connections")
  for c in connections:
    if connections[c]["srv"].connected:
      connections[c]["srv"].peer.disconnect()
    if connections[c]["cl"].connected:
      connections[c]["cl"].peer.disconnect()
#    OS.delay_msec(200)
    if srvhost != null:
      srvhost.unbind()
    if clhost != null:
      clhost.unbind()
    hosted = false

func _address(gdn_address):
 return [gdn_address.get_host(), gdn_address.get_port()]


func check_connection_state(cid, cnxtype="cl", connected=false):
  if not cid > 0 :
    return false
  if not typeof(cid) == TYPE_INT:
    g.report(["bad cid type", cid])
    return false
  var c = connections[cid][cnxtype]
  if connected:
    return c.connected
  return c.connected or c.initializing or c.fail


func check_connection(cid, cnxtype="cl"):
#  if cid > 3000:
  if cid > 100:
    return false
  return check_connection_state(cid, cnxtype, true)


func get_available_id():
  for i in range(2, maxplayer+1):
    if !check_connection(i):
      return i;
  return -1;


func get_cid_from_peer(peer, type="cl"):
  for c in connections:
#    if check_connection(i) and connections[i].peer.get_peer_id() == peer.get_peer_id():
    # address_cid.has(peer.get_address())
    if _address(peer.get_address()) == connections[c][type].address:
      return c;
  g.report("connection from unknown peer", [_address(peer.get_address())], 0, 4000)
  return -1;


func add_server_to_list(cid):
  if not servers.has(cid):
    servers.append(cid)


func get_token_cid(token):
  if token == 1:
    return 1
  if token_cid.has(token):
    return token_cid[token]
  return -1


func gen_token(srv,cid):
  var token = randi()
  cnx_tokens[srv][cid] = token
  return token


func add_expected_client(h, cid, token, src):
  g.report(["add expected client", h, cid, token], [], 0)
  token_cid[token] = cid
  connections[cid]["cl"].address.append(h)
  connections[cid]["cl"].token = token


func cnx_summary():
  var summ = {"servers":{}, "clients":{}}
  var cn
  for c in connections:
    for k in ["cl", "srv"]:
      var cat = "clients"
      if k == "srv":
        cat = "servers"
      if check_connection_state(c,k):
        cn = connections[c][k]
        summ[cat][c] = [cn.address]
        if cn.connected:
          summ[cat][c].append("connected")
        if cn.initializing:
          summ[cat][c].append("initializing")
        if cn.fail:
          summ[cat][c].append("fail")

#        summ["clients"][c] = [cn.address, 'connected', cn.connected, 'initializing', cn.initializing, 'fail', cn.fail]
#    if check_connection_state(c, "srv"):
#      cn = connections[c]["srv"]
      #summ["servers"][c] = [cn.address, 'connected', cn.connected, 'initializing', cn.initializing, 'fail', cn.fail]
  return summ

func reset_connetion(cid):
  g.report(["reset connection", cid], [connections[cid]["cl"], connections[cid]["srv"]], 0 ,0)
  connections[cid]["cl"] = NetActor.new();
  connections[cid]["srv"] = NetActor.new();

func start_server(ip="", port = 3000):
  if hosted:
    return;
  var address = GDNetAddress.new();
  address.set_host(ip);
  address.set_port(port);
  srvhost = GDNetHost.new();
  srvhost.set_max_channels(10)
  print("Strating server")
  # srvhost.set_max_bandwidth_out(3000)
  var err = srvhost.bind(address);
  if err:
    g.report(["Failed starting server", err, ip, port], [],0 ,0)
    print(err)
    srvhost.unbind()
    srvhost = null
    return false
  else:
    uid = 1
    servers.append(uid)
    hosted = true;
    is_server = true
    is_primary = true
    set_process(true);
    return true


func start_punch_proxy():
  start_proxy("", null,true)


func start_proxy(ip = "", port = 4000,punch = false):
  if srvhost:
    return
  var err
  if not punch:
    srvhost = GDNetHost.new()
    srvhost.set_max_channels(3)
    var address = GDNetAddress.new();
    address.set_host(ip);
    address.set_port(port);
    err =srvhost.bind(address);
  else:
    port = my_primary_addr[1]
    srvhost = clhost
    err = false
  if err:
    srvhost = null
    g.report(["start proxy failed"], [ip, port], 0 ,0)
#    print("Failed to start proxy: ", err)
    return
  else:
    rpc_id(primary_server, self, "register_server", [port])


func register_server(p, cid):
# called on primary server
# params [port, proxy client uid]
  g.report(["Actor trying to register as server", p, cid], [], 0)
  var h = connections[cid]["cl"].address[0]
  var can_connect =  connect_to_server(h, p, cid, 1)
  if can_connect:
    var current_cnx = {}
    cnx_tokens[cid] = {}
    for c in connections:
      if check_connection(c):
        var token = gen_token(cid,c)
        current_cnx[c] = [connections[c]["cl"].address[0], token]
    # complete server reg
    rpc_id(cid, self, "server_registered", [current_cnx])


func server_registered(current_cnx, src=1):
# called on proxy server
# params [cnx table, primary uid]
  is_server = true
  for c in current_cnx:
    add_expected_client(current_cnx[c][0], c, current_cnx[c][1], src)
  print("Server started")
  rpc_id(src, self, "server_started", [], {"force_proxy":1})


func server_started(cid=-1):
# called on primary to notify server is ready
# params [proxy uid]
  if not cnx_tokens.has(cid):
    print("proxy imposter fail")
    return
  var h = connections[cid]["srv"].address[0]
  var p = connections[cid]["srv"].address[1]
  add_server_to_list(cid)
#  ping(cid)
  for c in connections:
    if  not check_connection(c):
      continue
    rpc_id(c, self, "server_added", [h,p,cid, cnx_tokens[cid][c]])


func server_added(h, p, cid, token, src=1):
# called on all connected peers after new server is added
# params[proxy host, proxy port, proxy uid, connection token]
  print("server added notification", cid)
  if is_primary:
    return
  add_server_to_list(cid)
  if cid == uid:
    return
  connections[cid]["srv"].address = [h,p]
  connections[cid]["srv"].token = token
#  if not connection_inprogress:
  connect_next_server()
#  connect_to_server(h,p, cid, token)


func connect_to_server(ip = "localhost", port = 3000,cid=1, token=-1):
  print("Connecting to server:", ip, ",", port, ",", cid, ",", token)
  var address = GDNetAddress.new();
  address.set_host(ip);
  address.set_port(port);
  if not clhost:
    clhost = GDNetHost.new()
    clhost.set_max_channels(10)
    # clhost.set_max_bandwidth_out(3000)
    clhost.bind()
  var peer
  var connected = false;
  var attempts = 0;
  while !connected && attempts < 3:
    peer = clhost.connect(address, token);
    attempts += 1;
    for i in range(0,5):
      OS.delay_msec(100);
      if (clhost.is_event_available()):
        var event = clhost.get_event()
        if (event.get_event_type() == GDNetEvent.CONNECT):
          print("Connected.");
          connected = true;
          break;
  if !connected:
    print("Failed Connecting to ",ip,":",str(port),".");
    if cid == 1:
      clhost.unbind()
      clhost = null
    return false
  else:
    connections[cid]["srv"].peer = peer
    connections[cid]["srv"].address = [ip, port]
    connections[cid]["srv"].initializing = true
#    set_process(true);
    return true
# == Now waiting for server accept msg ==


func rpc_connect_servers(src=-1):
  initialized = true
  connect_next_server()


func connect_next_server(src=-1):
  if not initialized:
    connection_pending = true
    return
  if not connection_waiting.empty():
    var cw = connection_waiting
    if connections[cw[1]]["srv"].connected:
      connections[cw[1]]["srv"].initializing = false
      connection_waiting = []
#      connection_pending = false
    elif OS.get_ticks_msec() - connection_waiting[0] > 5000:
      connections[cw[1]]["srv"].fail = true
      connections[cw[1]]["srv"].initializing = false
      connection_waiting = []
#      connection_pending = false
    else:
      g.report(["---connect_next_server busy"], [connection_waiting], 0 ,500)
      connection_pending = true
      return
  for c in connections:
    var srv = connections[c]["srv"]
    if srv.connected or srv.address.empty() or srv.fail:
      continue
    if srv.initializing:
      g.report(["initializing connection", c], [srv.address], 0 ,1000)
      return
    var ok = connect_to_server(srv.address[0], srv.address[1], c, srv.token)
    if not ok:
      g.report(["connection failed", c], [], 0,0)
      srv.initializing = false
      srv.fail = true
      connection_waiting = []
      continue
    else:
      connection_waiting = [OS.get_ticks_msec(), c]
      connection_pending = true
      return
  if connection_pending:
    connection_pending = false
    connect_next_server()
  g.report(["++++++ done connecting"], [], 0, 5000)
  emit_signal("done_connecting")
#  connection_pending = false



func start_auto_proxy(force=false):
  if (is_proxy or force) and initialized and not is_server:
    g.report(["++Starting auto proxy"], [], 0 ,0)
    if not net.my_primary_addr.empty():
      var port
      if net.my_primary_addr[0] in ["", "127.0.0.1", "localhost", "0.0.0.0"]:
        randomize()
        port = int(rand_range(4000,5000))
      else:
        port = 4000
      net.start_proxy('', port)

func _process(delta):
  if connection_pending:
    g.report(["pending"], [],0,0)
    connect_next_server()
  check_events();
  if not process_data_queue.empty():
    for qd in process_data_queue:
      if g.t() > qd[-1]:
        var ds = qd[0]
        process_data(ds[0], ds[1])
        process_data_queue.erase(qd)
#        process_data_queue.pop_front()
  if not send_data_queue.empty():
    for qd in send_data_queue:
      if g.t() > qd[-1]:
        var ds = qd[0]
        ds.append(true) # skip_lag
        self.callv("send_data", ds)
#        send_data(ds[0], ds[1])
        send_data_queue.erase(qd)
#        process_data_queue.pop_front()
  if not replay_data.empty():
    if replay_state == REPLAY_PLAYING:
      var rd0 = replay_data[0]
      if g.tst() - replay_current_time > rd0[0]:
        g.report("rd0", rd0[1][0], 0 , 1000, {"only":false})
        replay_local_time = rd0[0]
        self.callv("process_data", [rd0[1][0], 0])
        replay_data.pop_front()
  elif replay_state == REPLAY_PLAYING:
    replay_state = REPLAY_STOPPED

func replay_init(data=null):
  if g.is_null(data):
    data = [] + data_log
  replay_data = data
  replay_current_time = g.t()
  replay_local_time = 0
  replay_state = REPLAY_PLAYING

func replay_pause():
  if replay_state == REPLAY_PLAYING:
    replay_state = REPLAY_PAUSED
    replay_pause_time = g.t()

func replay_resume():
  if replay_state == REPLAY_PAUSED:
    replay_state = REPLAY_PLAYING
    var time_elapsed_since_paused = g.t() - replay_pause_time
    replay_current_time += time_elapsed_since_paused


func check_events():
  if srvhost:
    while srvhost.is_event_available():
      process_server_event(srvhost.get_event())
  if clhost:
    while clhost.is_event_available():
      process_client_event(clhost.get_event());


func process_server_event(event):
  var peer = srvhost.get_peer(event.get_peer_id());
  var cid = get_cid_from_peer(peer)
  var token
  if event.get_event_type() == GDNetEvent.CONNECT:
    g.report(["Client trying to connect"], [_address(peer.get_address())], 0,0)
    var new = false
    if is_primary:
      new = true
      cid = get_available_id();
      g.report(["Primary assigning new id", cid], [], 0,0)
    else:
      token = event.get_data()
      cid = get_token_cid(token)
      if token != 1:
        g.report(["Secondary expecting token", token,connections[cid]["cl"].token,  cid], ['incoming host', peer.get_address().get_host(), 'saved host',connections[cid]["cl"].address[0]], 0)
      else:
        g.report(["Primary trying to connect to us"], [], 0,0)
    #if cid != -1 and (cid == 1 or new or connections[cid]["cl"].address[0] == peer.get_address().get_host()):
    if cid != -1 and ((cid == 1) or new or (token and (connections[cid]["cl"].token == token))):
      connections[cid]["cl"].connected = true;
      connections[cid]["cl"].peer = peer;
      connections[cid]["cl"].address = _address(peer.get_address());
      # send existing servers plus newly assigned id to client
      var srvlist = []
      if is_primary:
        for s in servers:
          if s == uid:
            continue
          var token = gen_token(s,cid)
          rpc_id(s,self, "add_expected_client", [connections[cid]["cl"].address[0], cid, token])
          srvlist.append([connections[s]["srv"].address,s, token])
      g.report(["Accepting connection", cid, connections[cid]["cl"].address], [], 0)
      peer.send_var([NET_ACCEPTED, cid, uid, connections[cid]["cl"].address, srvlist], GDNetMessage.RELIABLE, GDNetMessage.RELIABLE);
      emit_signal("client_connected", cid)
    return;
  elif event.get_event_type() == GDNetEvent.DISCONNECT:
    if is_server and cid != -1:
  #      var cid = get_cid_from_peer(peer);
#      if cid != -1:

      if is_primary:
#        connections[cid]["cl"] = NetActor.new();
#        connections[cid]["srv"] = NetActor.new();
        rpc_sync(self, "reset_connetion", [cid])
        emit_signal("client_disconnected", cid)
      else:
        connections[cid]["cl"].connected = false
        connections[cid]["srv"].connected = false
      peer = null;
      if cid == primary_server:
        print("Disconnected from primary server")
        reset()
        emit_signal("server_disconnected")
        return
      elif cid in servers:
        servers.erase(cid)
#        if not connection_inprogress:
        connect_next_server()
      print("Connection lost", cid);
      return
  elif (event.get_event_type() == GDNetEvent.RECEIVE):
    var channel = event.get_channel_id()
    g.report("receive channel", [event, channel], 3 , 1000, {"only":true})
    g.report("server receive", 5)
    var data = event.get_var();
    if is_server:
      # confirm src is really who it claims it is
      handle_server_receive(event,data, cid)
#    else:
#      process_data(data, cid)


func handle_server_receive(event,data, sender):
  g.report("server receieve event", [data, sender], 3)
#  var data = event.get_var();
  var target = data[1]
  if target == uid or target == 0:
    if event.get_channel_id() == 1:
#      g.chreport("channel 1", [data], 0 , 5000, {"only":true})
      process_data(data,sender)
    elif event.get_channel_id() == 4:
#      queue_data(data, sender, g.t() + 1000 + g.deep(sync_table, [sender,"latency"]))
      # queue_data(data, sender, g.t() + server_simulation_delay )
      queue_data(data, sender, g.t() )
    elif event.get_channel_id() == 5:
      gsvar.announce("thats", 1, "2")
      queue_data(data, sender, g.t() + 100 + 2*g.deep(sync_table, [sender,"latency"]) )
    else:
     process_data(data, sender)
      # var time = g.t() + server_simulation_delay - 2 * g.deep(sync_table, [sender,"latency"])
      # queue_data(data, sender,time)
  if target == uid:
    return
  # redirect data to target unless it was alreay redirected from a server
  var force_proxy = data[-1]
  if force_proxy == uid or not (sender in servers):
    # msg_type doubles as channel id
#    print(data.append)
    var msg_type = min(event.get_channel_id(), 2) # set reliabe for channel 2 or higher
    data[2] = sender
    # exclude sender
#    print(uid, " relaying ", sender, " data ", data)
    g.report(["relaying data from", sender], data, 1)
    send_data(data, msg_type, "cl", [sender])
    return


func process_client_event(event):
  g.report(["client event received",event.get_event_type()], [], 1 , 10000, {"only":false})
  var peer = clhost.get_peer(event.get_peer_id());
  var cid = get_cid_from_peer(peer, "srv")
  if event.get_event_type() == GDNetEvent.DISCONNECT:
    peer = null;
    if cid == primary_server:
      print("Disconnected from primary server")
      reset()
      emit_signal("server_disconnected")
      return
    servers.erase(cid)
    print("Server disconnected.");
#    peer = null;
    return
  elif (event.get_event_type() == GDNetEvent.RECEIVE):
    var data = event.get_var();
    if data[0] == NET_ACCEPTED:
      g.report(["NET_ACCEPTED", data], [], 0)
      process_net_accepted(peer, data)
      return
    process_data(data, cid)


func process_net_accepted(peer, data):
#      if not is_primary:
  var cid = data[2]
  if cid == primary_server:
    uid = data[1];
    my_primary_addr = data[3]
    servers = [1]
  connections[cid]["srv"].initializing = false
  connections[cid]["srv"].connected = true
  connections[cid]["srv"].peer = peer
  connections[cid]["srv"].address = _address(peer.get_address())
  print("Connection to server established")
  var srvs = data[4]
  for s in srvs:
    connections[s[1]]["srv"].address = s[0]
    connections[s[1]]["srv"].token = s[2]
#        connect_to_server(s[0][0], s[0][1], s[1], s[2])
  add_server_to_list(cid)
  ping(cid)
#      connection_waiting = []
#      if initialized:
#        connect_next_server()
  emit_signal("connected_to_server", cid)


func short_path(lp):
  if typeof(lp) == 16:
    return String(lp).hash()
  elif typeof(lp) == 2:
    return lp
  else:
    return null
#  if lp.begin_with("/root/world/env/players"):
#    return [1, int(lp.split("/").back())]

func long_path(sp):
#  var lp = path_ht[sp[0]]
#  if sp.length() > 1:
#    lp = lp + "/" + String(sp[1])
  if typeof(sp) == 2:
    if path_ht.has(sp):
      return path_ht[sp]
    else:
      return null
  return sp

func queue_data(data, sender, time):
#  process_data_queue.append([[data,sender], g.t() + 1000 - g.deep(sync_table, [sender,"latency"])])
  process_data_queue.append([[data,sender], time])

func process_data(data, sender):
  g.report(["process data", data[1], data[4]], [data, net.path_ht], 5)
  g.report("report process data", [data,sender], 3 , 0, {"only":true})
#  if not has_node("/root/world/env/players/1"):
#    return
#  data[3] = path_ht[data[3]
  var lp = long_path(data[3])
  if g.is_null(lp):
    g.report("bad short path", [data], 0 , 1000, {"only":false})
    return
#  g.report("process data path", [lp], 0 , 0, {"only":false})
  data[3] = lp
  if data[0] == NET_RPC:
    process_rpc_data(data, sender)
  if data[0] == NET_RSET:
    process_rset_data(data, sender)


func process_rpc_data(data, sender):
  g.report("received rpc data", [data, sender], 2 , -1, {"only":false})
  var src
  if data[2]:
    src = data[2]
  else:
    src = sender
  var path = data[3]
  var fname = data[4]
  var args = data[5] + [src]
  var node = get_node(path)
  if node: # errchck
    if node.is_queued_for_deletion():
      g.log2f(str(["catch the bad bouy", node.get_path()]))
      return
#    g.report(["rec rpc", fname], [path, fname, args],0 ,-1)
    node.callv(fname, args)


func process_rset_data(data,sender):
  var src
  if data[2]:
    src = data[2]
  else:
    src = sender
  var path = data[3]
  var vname  = data[4];
  var node = get_node(path)
#  if check_perms(node,["master"], src):
  if node:
    node.set(vname, data[5])
  else:
    g.report("!!Node missing in rset_data", [path], 0)


func get_target_latency(target):
  g.report("target+uid", [target, uid], 3 , 0, {"only":true})
  if target == uid or target == 0 or not sync_table.has(target):
    return 0
  var player_sync = sync_table[target]
  if player_sync.has("latency"):
    return player_sync["latency"]


func send_data(data, msg_type, target_type = "cl", excl=[], force_proxy=-1, short_path=false, channel=0, skip_lag=false):
  g.scrn_dbg("send_data_vvv", data)
  g.report("data", [data], 3 , 0, {"only":true})
  var target = data[1]
#   if is_primary and reverse_lag and target != 0 and not skip_lag:
#     var latency = get_target_latency(target)
#     if g.is_null(latency):
#       latency = 0
# #    var _short_path = false
#     send_data_queue.append([[data, msg_type, target_type, excl, force_proxy, short_path, channel], g.t() + latency])
#     return
  # if is_primary and target != 0 and not skip_lag:
  # if target != 0 and not skip_lag:
  if  not skip_lag:
    var latency = server_simulation_delay
    if is_primary:
      latency = server_simulation_delay
    if g.is_null(latency):
      latency = 0
#    var _short_path = false
    send_data_queue.append([[data, msg_type, target_type, excl, force_proxy, short_path, channel], g.t() + latency])
    return
  if not setup_complete:
    return
  var use_proxy = proxy
  if force_proxy > 0:
    use_proxy = force_proxy
#  var target = data[1]
  if target == net.uid:
    g.report(["data sent to self"], [data],10000,0 )
    return
#  data[3] = String(data[3]).hash()
  if short_path:
#    g.report("data 3", [data[3], short_path(data[3])], 0 , 0, {"only":false})
    data[3] = short_path(data[3])
  if target== 0:
    pass
#    data_log.append([g.t(), [data, msg_type, target_type, excl, force_proxy, short_path, channel]])
  if force_proxy > 0 or not is_server:
    if !use_proxy || !check_connection(use_proxy, "srv") || connections[use_proxy]["srv"].peer == null:
      g.report(["!!send data no useable proxy, revert to 1"], [], 0)
      use_proxy = 1
    data.append(force_proxy)
    if target == 0:
      if channel == 4:
        g.report("chr send from client", [data, msg_type], 0 , 0, {"only":false})
      connections[use_proxy]["srv"].peer.send_var(data, channel, msg_type);
      # == send directly to servers ==
      for s in servers:
        if s == use_proxy:
          continue
        data[1] = s
        connections[s]["srv"].peer.send_var(data, channel, msg_type);
    elif target in servers:
      connections[target]["srv"].peer.send_var(data, channel, msg_type);
    else:
      if not g.is_null(connections[use_proxy]["srv"].peer):
        connections[use_proxy]["srv"].peer.send_var(data, channel, msg_type);
  else:
    if target == 0:
      if channel == 4:
        g.report("ch4 data from server", [data,msg_type], 0 , 0, {"only":false})
        # remember to uncomment
      data_log.append([g.t(), [data, msg_type, target_type, excl, force_proxy, short_path, channel]])
      for c in connections:
        # == do not relay to servers or excluded or self  ==
        if (data[2] && c in servers) || c in excl || c == uid:
          continue
        if !check_connection(c, target_type)  || connections[c][target_type].peer == null:
#          g.report(["!!Trying to send to unreachable actor", c], [], 0)
          continue;
        if reverse_lag:
          var target_data = [] + data
          target_data[1] = c
#          g.report("mangled data", [String(target_data[3]).hash()], 0 , 0, {"only":false})
          send_data(target_data, msg_type, target_type, excl, force_proxy, short_path, channel)
        else:
          connections[c][target_type].peer.send_var(data, channel, msg_type);
    elif target == uid:
      pass
    else:
      if check_connection(target, target_type):
        connections[target][target_type].peer.send_var(data, channel, msg_type);
      else:
        g.report(["!!Trying to send to unreachable actor", target],[], 0)


func rset(where, name, value, params = {}):
#  g.report(["rset call", name],[], 5)
  g.report("rset", [where, name,value], 5 , -1, {"only":false})
  var defaults = {'reliable':true, 'target':0, 'excl':[], 'force_proxy':-1, 'short_path':false, 'channel':-1}
  for k in defaults:
    if not params.has(k): params[k] = defaults[k]
  # var msg_type = GDNetMessage.UNSEQUENCED;
  var msg_type = GDNetMessage.SEQUENCED;
  if params["reliable"]:
    msg_type = GDNetMessage.RELIABLE;
  if params["channel"] == -1:
    params["channel"] = msg_type
  var path = where.get_path()
  where.set(name, value)
  var target_type
#  if name =="sup":
#    g.report("sup where", [path, short_path(path)], 0 , 1000, {"only":false})
  send_data([NET_RSET, params['target'],null, path, name, value ], msg_type, "cl", params['excl'], params['force_proxy'],params['short_path'], params["channel"]);


func rpc(where, fname, args, params = {}):
  g.report(["rpc call", fname],[], 3)
  var defaults = {'reliable':true, 'target':0, 'excl':[], 'force_proxy':-1, 'short_path':false, 'channel':-1}
  for k in defaults:
    if not params.has(k): params[k] = defaults[k]
  # var msg_type = GDNetMessage.UNSEQUENCED;
  var msg_type = GDNetMessage.SEQUENCED;
  if params["reliable"]:
    msg_type = GDNetMessage.RELIABLE;
  if params["channel"] == -1:
    params["channel"] = msg_type
  var path = where.get_path()
  send_data([NET_RPC,params['target'],null , path, fname, args  ], msg_type,"cl", params['excl'], params['force_proxy'],params["short_path"] ,params["channel"]);

func rpc_id(id, where, fname, args, params={}):
  params["target"] = id
  rpc(where, fname, args, params)

func rpc_sync(where, fname, args, params = {}):
  where.callv(fname, args)
  rpc(where, fname, args, params)


func rpc_sync_id(id, where, fname, args, params={}):
  params["target"] = id
  rpc_sync(where, fname, args, params )


func rpc_sync_server(where, fname, args, params = {}):
  params["target"] = primary_server
  rpc_sync(where, fname, args, params)


func check_perms(node, perms, cid):
  if cid == -1:
    return true
  for p in perms:
    if p == "anyone":
     return true
    elif p == "primary":
      if cid != primary_server:
        return false
    elif p == "server":
     if not cid in servers:
       return false
    elif p == "master":
     if not node.get_path() in  master_nodes[cid]:
       return false
  return true

func ping_all(src=0):
  for i in get_node("/root/world/env/players").get_children():
    ping(i.get_name())

func ping(target, src=0):
  target = int(target)
  var channel = 1
  var reliable = true
  if not sync_table.has(target):
    sync_table[target] = {}

  # direct ping (through target)
  if target in net.servers:
    var direct_ping_id = randi()
    pings[direct_ping_id] = {"target":target, "sent_at":OS.get_ticks_msec(), "direct":true}
#    global.report(["++ping target",direct_ping_id, pings[direct_ping_id]], [],3)
    net.rpc_id(target,self, "ping_respond", [direct_ping_id], {"force_proxy":target, "reliable":reliable,"channel":channel} )
  if target == net.uid:
    return
  var ping_id = randi()
  # normal ping (through default proxy)
  pings[ping_id] = {"target":target, "sent_at":OS.get_ticks_msec(), "direct":false}
#  global.report(["++ping target",ping_id,pings[ping_id]], [],3)
  net.rpc_id(target,self, "ping_respond", [ping_id], {"reliable":reliable,"channel":channel} )


func ping_respond(ping_id, src=0):
  var channel = 1
  var reliable = true
#  global.report(["--ping respond", ping_id], [src],3)
  var sender = src
  net.rpc_id(sender, self, "ping_rsp_handle", [ping_id, OS.get_ticks_msec()], { "reliable":reliable, "channel":channel})

func ping_rsp_handle(ping_id, remote_receive_at, src=0):
#  global.report(["++ping handle response", ping_id], [src],3)
  var target = src
  var return_at = OS.get_ticks_msec()
  var sent_at = pings[ping_id]["sent_at"]
  var latency = (return_at - sent_at) / 2
#  var latency = (return_at - sent_at)
  var diff
  var test = "latency"
  var direct = pings[ping_id]["direct"]
  if direct:
    test = "direct_latency"
#  g.report("update " + test , [], 3)

  var old_latency = g.deep(sync_table, [target,test])
  if old_latency:
    diff = latency - old_latency

  sync_table[target][test] = latency
  sync_table[target]["rel_ticks"] = sent_at - (remote_receive_at - latency)

  emit_signal("latency_change", target, diff, direct)
  pings.erase(ping_id)